﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Xml.Linq;

namespace RH.ImportGoogleAnalytics.Analytics
{
	/// <summary>
	/// thx to Jacob Reimers
	/// http://www.reimers.dk/blogs/jacob_reimers_weblog/archive/2009/05/09/added-google-analytics-reader-for-net.aspx
	/// </summary>
	public class ReportRequestor
	{
		private const string AuthUrlFormat = "accountType=GOOGLE&Email={0}&Passwd={1}&source=reimers.dk-analyticsreader-0.1&service=analytics";
		private const string RequestUrlFormat = "https://www.google.com/analytics/feeds/data?ids={0}&dimensions={1}&metrics={2}&start-date={3}&end-date={4}&itemsPerPage=1000&max-results=1000&start-index={5}";
		private string _Password;

		private string _Token;

		private string _Username;

		public ReportRequestor(string email, string password)
		{
			_Username = email;
			_Password = password;
		}

		public string Email
		{
			get { return _Username; }
			set
			{
				if (string.Equals(_Username, value)) return;
				_Username = value;
				_Token = null;
			}
		}

		public string Password
		{
			get { return _Password; }

			set
			{
				if (string.Equals(_Password, value)) return;
				_Password = value;

				_Token = null;
			}
		}

		#region Methods
		private string GetToken(string username, string password)
		{
			if (string.IsNullOrEmpty(_Username) || string.IsNullOrEmpty(_Password))
			{
				throw new ArgumentNullException("Username, Password", "Username and/or password not set");
			}

			string authBody = string.Format(AuthUrlFormat, username, password);
			var req = (HttpWebRequest) WebRequest.Create("https://www.google.com/accounts/ClientLogin");
			req.Method = "POST";
			req.ContentType = "application/x-www-form-urlencoded";
			req.UserAgent = "Hezser.de req";

			using (Stream stream = req.GetRequestStream())
			using (var sw = new StreamWriter(stream))
			{
				sw.Write(authBody);
				sw.Close();
			}

			using (var response = (HttpWebResponse) req.GetResponse())
			using (var sr = new StreamReader(response.GetResponseStream()))
			{
				string token = sr.ReadToEnd();
				string[] tokens = token.Split(new[] {"\n"}, StringSplitOptions.RemoveEmptyEntries);

				foreach (string item in tokens)
				{
					if (item.StartsWith("Auth="))
					{
						return item.Replace("Auth=", "");
					}
				}
			}
			return string.Empty;
		}

		public IEnumerable<AnalyticsAccountInfo> GetAccounts()
		{
			if (string.IsNullOrEmpty(_Token))
			{
				_Token = GetToken(_Username, _Password);
			}

			var req = (HttpWebRequest) WebRequest.Create("https://www.google.com/analytics/feeds/accounts/default");
         req.Headers.Add("Authorization: GoogleLogin auth=" + _Token);

			var response = (HttpWebResponse) req.GetResponse();
         Stream responseStream = response.GetResponseStream();
         var sr = new StreamReader(responseStream);

			string responseXml = sr.ReadToEnd();
         XDocument doc = XDocument.Parse(responseXml);
         XNamespace dxpSpace = doc.Root.GetNamespaceOfPrefix("dxp");
         XNamespace defaultSpace = doc.Root.GetDefaultNamespace();
         
			var entries = from en in doc.Root.Descendants(defaultSpace + "entry")
			              select new AnalyticsAccountInfo

			                     	{
			                     		AccountID = en.Elements(dxpSpace + "property").Where(xe => xe.Attribute("name").Value == "ga:accountId").First().Attribute("value").Value,
			                     		AccountName = en.Elements(dxpSpace + "property").Where(xe => xe.Attribute("name").Value == "ga:accountName").First().Attribute("value").Value,
			                     		ID = en.Element(defaultSpace + "id").Value,
			                     		Title = en.Element(defaultSpace + "title").Value,
			                     		ProfileID = en.Elements(dxpSpace + "property").Where(xe => xe.Attribute("name").Value == "ga:profileId").First().Attribute("value").Value,
			                     		WebPropertyID = en.Elements(dxpSpace + "property").Where(xe => xe.Attribute("name").Value == "ga:webPropertyId").First().Attribute("value").Value
			                     	};

			return entries;
		}

		private XDocument GetReport(AnalyticsAccountInfo account, IEnumerable<Dimension> dimensions, IEnumerable<Metric> metrics, DateTime from, DateTime to, int startIndex)
		{
			if (string.IsNullOrEmpty(_Token))
			{
				_Token = GetToken(_Username, _Password);
			}

			var dims = new StringBuilder();
         foreach (Dimension dimension in dimensions)
			{
				dims.AppendFormat("ga:{0},", dimension);
			}
			string dimensionsParameter = dims.ToString().Trim(new[] { ',' });

			var mets = new StringBuilder();
         foreach (Metric metric in metrics)
			{
				mets.AppendFormat("ga:{0},", metric);
			}
			string metricsParameter = mets.ToString().Trim(new[] { ',' });

			string requestUrl = string.Format(RequestUrlFormat,
				"ga:" + account.ProfileID, dimensionsParameter, metricsParameter, from.ToString("yyyy-MM-dd"), to.ToString("yyyy-MM-dd"), startIndex);
         var req = (HttpWebRequest) WebRequest.Create(requestUrl);
         req.Headers.Add("Authorization: GoogleLogin auth=" + _Token);

			var response = (HttpWebResponse) req.GetResponse();
         Stream responseStream = response.GetResponseStream();
         string responseXml = new StreamReader(responseStream, Encoding.UTF8, true).ReadToEnd();
         XDocument doc = XDocument.Parse(responseXml);

			return doc;
		}

		public IEnumerable<GenericEntry> RequestReport(AnalyticsAccountInfo account, IEnumerable<Dimension> dimensions, IEnumerable<Metric> metrics, DateTime from, DateTime to)
		{
			var genericEntries = new List<GenericEntry>();
         IEnumerable<GenericEntry> currentEntries;
			int startindex = 1;

			do
			{
				XDocument doc = GetReport(account, dimensions, metrics, from, to, startindex);
				XNamespace dxpSpace = doc.Root.GetNamespaceOfPrefix("dxp");
				XNamespace defaultSpace = doc.Root.GetDefaultNamespace();


				currentEntries = (from r in doc.Root.Descendants(defaultSpace + "entry")
				                  select new GenericEntry

				                         	{
				                         		Dimensions = new List<KeyValuePair<Dimension, string>>(
				                         			from rd in r.Elements(dxpSpace + "dimension")
				                         			select new KeyValuePair<Dimension, string>(
				                         				(Dimension) Enum.Parse(typeof (Dimension), rd.Attribute("name").Value.Replace("ga:", ""), true),
				                         				rd.Attribute("value").Value)),
				                         		Metrics = new List<KeyValuePair<Metric, string>>(
				                         			from rm in r.Elements(dxpSpace + "metric")
				                         			select new KeyValuePair<Metric, string>(
				                         				(Metric) Enum.Parse(typeof (Metric), rm.Attribute("name").Value.Replace("ga:", ""), true),
				                         				rm.Attribute("value").Value))
				                         	});
				genericEntries.AddRange(currentEntries);
				startindex = startindex + 1000;

				// repeat until we get less than the maximum rows for one request
			} while (currentEntries.Count() == 1000);

			return genericEntries;
		}

		/// <summary>
		/// get data ordered by date and country
		/// </summary>
		/// <param name="account"></param>
		/// <param name="from"></param>
		/// <param name="to"></param>
		/// <returns></returns>
		public IEnumerable<CityReport> GetUserCountByCountry(AnalyticsAccountInfo account, DateTime from, DateTime to)
		{
			IEnumerable<GenericEntry> report = RequestReport(account, new[] { Dimension.Country, Dimension.Date }, new[] { Metric.Visits }, from, to);

			var cr = from r in report
						select new CityReport
                  (
							r.Dimensions.First(d => d.Key == Dimension.Date).Value,
							r.Dimensions.First(d => d.Key == Dimension.Country).Value,
							Convert.ToInt32(r.Metrics.First(m => m.Key == Metric.Visits).Value)
						);

			return cr.ToList().OrderBy(r => r.Date).ThenBy(r => r.Country);
		}
		#endregion
	}
}