﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.SharePoint;
using RH.ImportGoogleAnalytics.Analytics;
using Microsoft.SharePoint.Utilities;

namespace RH.ImportGoogleAnalytics
{
	class Program
	{
		static void Main(string[] args)
		{
			SPWeb web = null;
			try
			{
				Hashtable arguments = ArgumentParser.Parse(args);
				if (arguments.ContainsKey("?") || arguments.ContainsKey("Invalid") || (arguments.Count != 5 && arguments.Count != 6) ||
				    !arguments.ContainsKey("WebUrl") ||!arguments.ContainsKey("ListName") ||!arguments.ContainsKey("GoogleLoginName") ||!arguments.ContainsKey("GooglePassword") ||!arguments.ContainsKey("GoogleTitle") ||
				    (arguments.Count == 6 && !arguments.ContainsKey("StartDate")))
				{
					Console.WriteLine(Constants.Usage());
				}
				else
				{
					// get sharepoint objects
					Console.Write("Connecting to SPWeb...");
					web = GetWeb(arguments);
					Console.WriteLine(" Done.");
					Console.Write("Getting list...");
					SPList list = EnsureStatisticsList(web, arguments);
					Console.WriteLine(" Done.");

					// get data from google
					Console.Write("Retrieving data from Google Analytics...");
					IEnumerable<CityReport> analyticsData = GetDateFromGoogle(arguments);
					Console.WriteLine(" {0} rows received.", analyticsData.Count());

					// store data in list
					web.AllowUnsafeUpdates = true;
					InsertDataToList(web, list, analyticsData);
					Console.WriteLine("Finished.");
				}
			}
			catch (Exception ex)
			{
				if (web != null)
				{
					web.AllowUnsafeUpdates = false;
					web.Dispose();
				}
				Console.WriteLine(Environment.NewLine + "Exception:");
				Console.WriteLine(ex.Message);
				Console.WriteLine(ex.StackTrace);
			}
		}

		/// <summary>
		/// open SPWeb from passed Url
		/// </summary>
		/// <param name="arguments"></param>
		/// <returns></returns>
		private static SPWeb GetWeb(IDictionary arguments)
		{
			try
			{
				using (var site = new SPSite(arguments["WebUrl"].ToString()))
				{
					return site.OpenWeb();
				}
			}
			catch (Exception ex)
			{
				throw new InvalidOperationException("Could not get web from url: " + ex.Message);
			}
		}

		/// <summary>
		/// Ensure list with all needed fields
		/// </summary>
		/// <param name="web"></param>
		/// <param name="arguments"></param>
		private static SPList EnsureStatisticsList(SPWeb web, IDictionary arguments)
		{
			try
			{
				web.AllowUnsafeUpdates = true;
				SPListCollection lists = web.Lists;
				string listName = arguments["ListName"].ToString();

				// ensure list
				if (lists.Cast<SPList>().Any(l => l.Title == listName) == false)
				{
					// create the list
					Console.WriteLine("The list '{0}' does not exist. It will be created.", listName);
					lists.Add(listName, string.Empty, SPListTemplateType.GenericList);
				}
				SPList list = lists.Cast<SPList>().SingleOrDefault(l => l.Title == listName);

				// ensure fields
				SPFieldCollection fields = list.Fields;
				SPField title = fields[SPBuiltInFieldId.Title];
				title.Title = Constants.FieldTitleDisplay;
				title.Update();

				SPField date;
				SPView defaultView = list.Views[list.DefaultView.ID];
				if (!fields.GetField(Constants.FieldDate, out date))
				{
					Console.WriteLine("The list '{0}' does not contain a field '{1}. It will be created.", listName, Constants.FieldDate);
					fields.Add(Constants.FieldDate, SPFieldType.DateTime, true);
					date = fields.GetFieldByInternalName(Constants.FieldDate);
					var dateField = (SPFieldDateTime)date;
					dateField.DisplayFormat = SPDateTimeFieldFormatType.DateOnly;
					dateField.Update();

					defaultView.ViewFields.Add(date.Title);
					defaultView.Update();
				}

				SPField count;
				if (!fields.GetField(Constants.FieldCount, out count))
				{
					Console.WriteLine("The list '{0}' does not contain a field '{1}. It will be created.", listName, Constants.FieldCount);
					fields.Add(Constants.FieldCount, SPFieldType.Integer, true);
					count = fields.GetFieldByInternalName(Constants.FieldCount);
					count.Update();

					defaultView.ViewFields.Add(count.Title);
					defaultView.Update();
				}
				return list;
			}
			finally
			{
				web.AllowUnsafeUpdates = false;
			}
		}

		private static IEnumerable<CityReport> GetDateFromGoogle(IDictionary arguments)
		{
			// generate request
			var requestor = new ReportRequestor(arguments["GoogleLoginName"].ToString(), arguments["GooglePassword"].ToString());

			// get data for specific website
			IEnumerable<AnalyticsAccountInfo> accounts = requestor.GetAccounts();
			AnalyticsAccountInfo account = accounts.First(a => a.Title == arguments["GoogleTitle"].ToString());

			DateTime startDate = DateTime.Today.AddDays(-1);
			if (arguments.Contains("StartDate"))
			{
				DateTime.TryParse(arguments["StartDate"].ToString(), out startDate);
			}
			DateTime endDate = DateTime.Today.AddSeconds(-1);
			IEnumerable<CityReport> googleData = requestor.GetUserCountByCountry(account, startDate, endDate);
			return googleData;
		}

		private static void InsertDataToList(SPWeb web, SPList list, IEnumerable<CityReport> analyticsData)
		{
			IEnumerable<IGrouping<DateTime, CityReport>> groupedByDate = analyticsData.GroupBy(d => d.Date);
			// work on each day
			foreach (IGrouping<DateTime, CityReport> dataByDay in groupedByDate)
			{
				Console.WriteLine("Working on day {0}. Found {1} items to insert..", dataByDay.Key, dataByDay.Count());
				// delete existing listitems for this day
				DeleteItemsForDay(web, list, dataByDay.First().Date);

				// insert items into list
				InsertItemsForDay(web, list, dataByDay);
			}
		}

		private static void InsertItemsForDay(SPWeb web, SPList list, IEnumerable<CityReport> dataByDay)
		{
			string insertResult = string.Empty;
			try
			{
				string listId = list.ID.ToString("B");
				int index = 1;
				var addBuilder = new StringBuilder();
				addBuilder.Append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
				addBuilder.Append("<Batch OnError=\"Return\">");

				foreach (CityReport data in dataByDay)
				{
					// work on each data element
					addBuilder.AppendFormat("<Method ID=\"{0}\">", index);
					addBuilder.AppendFormat("<SetList Scope='Request'>{0}</SetList>", listId);
					addBuilder.Append("<SetVar Name=\"ID\">New</SetVar>");
					addBuilder.Append("<SetVar Name=\"Cmd\">Save</SetVar>");
					addBuilder.AppendFormat("<SetVar Name=\"urn:schemas-microsoft-com:office:office#Title\">{0}</SetVar>", data.Country);
					addBuilder.AppendFormat("<SetVar Name=\"urn:schemas-microsoft-com:office:office#{0}\">{1}</SetVar>", Constants.FieldDate, SPUtility.CreateISO8601DateTimeFromSystemDateTime(data.Date));
					addBuilder.AppendFormat("<SetVar Name=\"urn:schemas-microsoft-com:office:office#{0}\">{1}</SetVar>", Constants.FieldCount, data.Count);
					addBuilder.Append("</Method>");

					index++;
				}
			
				addBuilder.Append("</Batch>");

				Console.WriteLine("\tInserting data into list...");
				insertResult = web.ProcessBatchData(addBuilder.ToString());
			}
			catch (Exception ex)
			{
				throw new SPException(ex.Message + Environment.NewLine + insertResult);
			}
		}

		/// <summary>
		/// delete all listitems which match the date
		/// </summary>
		/// <param name="web"></param>
		/// <param name="list"></param>
		/// <param name="date"></param>
		private static void DeleteItemsForDay(SPWeb web, SPList list, DateTime date)
		{
			string deleteResult = string.Empty;
			try
			{
				var query = new SPQuery
				            	{
				            		Query = string.Format("<Where><Eq><FieldRef Name='Date' /><Value IncludeTimeValue='FALSE' Type='DateTime'>{0}</Value></Eq></Where>",
				            		                      SPUtility.CreateISO8601DateTimeFromSystemDateTime(date))
				            	};
				// get ids for items which need to be deleted
				IEnumerable<int> items = list.GetItems(query).Cast<SPListItem>().Select(i => i.ID);
				if (items.Count() == 0)
				{
					return;
				}
				Console.WriteLine("\tDeleting existing items for the day...");

				// build batchupdate string
				var sbDelete = new StringBuilder();
				sbDelete.Append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
				sbDelete.Append("<Batch>");
				foreach (int id in items)
				{
					sbDelete.Append("<Method>");
					sbDelete.AppendFormat("<SetList Scope=\"Request\">{0}</SetList>", list.ID);
					sbDelete.AppendFormat("<SetVar Name=\"ID\">{0}</SetVar>", id);
					sbDelete.Append("<SetVar Name=\"Cmd\">Delete</SetVar>");
					sbDelete.Append("</Method>");
				}
				sbDelete.Append("</Batch>");

				deleteResult = web.ProcessBatchData(sbDelete.ToString());
			}
			catch (Exception ex)
			{
				throw new SPException(ex.Message + Environment.NewLine + deleteResult);
			}
		}
	}
}
